

import os
import tempfile
import unittest

import dendropy
import numpy as np
from dendropy.simulate import treesim


def add_trifurication(tree):
    parent_node = list(tree.leaf_node_iter())[0].parent_node

    t1 = dendropy.Taxon(f'X1')
    t2 = dendropy.Taxon(f'X2')
    t3 = dendropy.Taxon(f'X3')

    tree.taxon_namespace.add_taxon(t1)
    tree.taxon_namespace.add_taxon(t2)
    tree.taxon_namespace.add_taxon(t3)

    child_a = dendropy.Node(edge_length=1.234)
    child_b = dendropy.Node(edge_length=1.234)
    child_c = dendropy.Node(edge_length=4.123)

    child_a.taxon = t1
    child_b.taxon = t2
    child_c.taxon = t3

    parent_node.add_child(child_a)
    parent_node.add_child(child_b)
    parent_node.add_child(child_c)


def get_test_tree(n: int, trifurication=False) -> dict:
    tree = treesim.birth_death_tree(birth_rate=1.0, death_rate=0.5, num_extant_tips=n)

    if trifurication:
        add_trifurication(tree)

    for i, edge in enumerate(tree.preorder_node_iter()):
        edge.edge_length = i

    n_taxa = len(tree.taxon_namespace)
    pdm = tree.phylogenetic_distance_matrix()
    taxa = sorted(pdm.taxon_iter())
    pd_mat = np.zeros((n_taxa, n_taxa))
    pd_mat_norm = np.zeros((n_taxa, n_taxa))
    nd_mat = np.zeros((n_taxa, n_taxa))
    nd_mat_norm = np.zeros((n_taxa, n_taxa))
    for i, t1 in enumerate(taxa):
        for j, t2 in enumerate(taxa):
            pd = pdm.patristic_distance(t1, t2)
            pd_norm = pdm.patristic_distance(t1, t2, is_normalize_by_tree_size=True)
            nd = pdm.path_edge_count(t1, t2)
            nd_norm = pdm.path_edge_count(t1, t2, is_normalize_by_tree_size=True)
            pd_mat[i, j] = pd
            pd_mat_norm[i, j] = pd_norm
            nd_mat[i, j] = nd
            nd_mat_norm[i, j] = nd_norm

    return {'tree': tree,
            'length': tree.length(),
            'taxa': tuple([x.label for x in taxa]),
            'pd_mat': pd_mat,
            'pd_mat_norm': pd_mat_norm,
            'nd_mat': nd_mat,
            'nd_mat_norm': nd_mat_norm}

def main():
    N = 10

    test_tree = get_test_tree(N, trifurication=False)

    print(test_tree['tree'].as_string(schema='newick'))
    print(test_tree['taxa'])

    for i in range(N):
        for j in range(i+1):
            cur_val = test_tree['pd_mat'][i, j]
            print(f'assert_eq!(arr[[{i}, {j}]], {cur_val});')

    print('---')

    for i in range(N):
        for j in range(i+1):
            cur_val = test_tree['pd_mat_norm'][i, j]
            print(f'assert_eq!(arr[[{i}, {j}]], {cur_val});')

    return

if __name__ == '__main__':
    main()


class TestPhyloDM(unittest.TestCase):

    def test_add_node(self):
        pdm = PhyloDM()
        pdm.add_node(taxon=None)
        pdm.add_node(taxon='b')

        nodes = pdm.get_nodes()
        self.assertEqual(len(nodes), 2)

    def test_add_edge(self):
        pdm = PhyloDM()
        a = pdm.add_node(taxon=None)
        b = pdm.add_node(taxon='b')
        pdm.add_edge(a, b, length=0.2)

    def test_dm(self):
        test_tree = get_test_tree(50)
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_path = os.path.join(tmpdir, 'test.tree')
            with open(tmp_path, 'w') as f:
                f.write(test_tree['tree'].as_string(schema='newick')[5:])

            pdm = PhyloDM.load_from_newick_path(tmp_path)

        dm = pdm.dm(norm=False)
        delta = test_tree['pd_mat'] - dm

        self.assertTrue(np.all(delta < 1e-6))
        self.assertAlmostEqual(pdm.length(), test_tree['length'], places=6)
        self.assertTrue(test_tree['taxa'] == tuple(pdm.taxa()))

    def test_dm_norm(self):
        test_tree = get_test_tree(50)
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_path = os.path.join(tmpdir, 'test.tree')
            with open(tmp_path, 'w') as f:
                f.write(test_tree['tree'].as_string(schema='newick')[5:])

            pdm = PhyloDM.load_from_newick_path(tmp_path)

        dm_norm = pdm.dm(norm=True)
        delta_norm = test_tree['pd_mat'] / pdm.length() - dm_norm

        self.assertTrue(np.all(delta_norm < 1e-6))
        self.assertAlmostEqual(pdm.length(), test_tree['length'], places=6)
        self.assertTrue(test_tree['taxa'] == tuple(pdm.taxa()))

    def test_load_from_dendropy(self):
        test_tree = get_test_tree(50)

        pdm = PhyloDM.load_from_dendropy(test_tree['tree'])
        dm = pdm.dm(norm=False)
        delta = test_tree['pd_mat'] - dm

        print(test_tree['pd_mat'])
        print(dm)
        print('--')
        print(delta)

        self.assertTrue(np.all(delta < 1e-6))
        self.assertAlmostEqual(pdm.length(), test_tree['length'], places=6)
        return

    def test_load_from_dendropy_with_trifurication(self):
        test_tree = get_test_tree(50, trifurication=True)

        pdm = PhyloDM.load_from_dendropy(test_tree['tree'])
        dm = pdm.dm(norm=False)
        delta = test_tree['pd_mat'] - dm

        self.assertTrue(np.all(delta < 1e-6))
        self.assertAlmostEqual(pdm.length(), test_tree['length'], places=6)
        self.assertTrue(test_tree['taxa'] == tuple(pdm.taxa()))
        return



import random

import typer
from dendropy.simulate import treesim


def main(n_taxa: int, output_path: str):
    tree = treesim.birth_death_tree(
        birth_rate=1.0,
        death_rate=0.2,
        num_extant_tips=n_taxa,
        rng=random.Random(42)
    )
    with open(output_path, 'w') as f:
        f.write(tree.as_string(schema='newick')[5:])


if __name__ == "__main__":
    typer.run(main)



from __future__ import annotations

from typing import Optional, List

import dendropy
import numpy as np

from .pdm import PhyloDM as PDM


class PhyloDM:

    def __init__(self):
        """Initialize the PhyloDM object."""
        self._rs = PDM()

    @classmethod
    def load_from_newick_path(cls, path: str) -> 'PhyloDM':
        """Load a tree from a Newick file.

        Args:
            path: The path to the Newick file.
        """
        try:
            pdm = cls()
            pdm._rs.load_from_newick_path(path=path)
            return pdm
        except Exception as e:
            print(f'Unable to load newick tree using light_phylogeny (Rust). '
                  f'This is likely due to it not supporting the extended Newick format... '
                  f'falling back to DendroPy to load the tree.')
            tree = dendropy.Tree.get(path=path, schema='newick')
            return cls.load_from_dendropy(tree)

    @classmethod
    def load_from_dendropy(cls, tree: dendropy.Tree) -> 'PhyloDM':
        """Load a tree from a Dendropy tree object.

        Args:
            tree: The Dendropy tree object.
        """
        pdm = cls()

        node_to_id = dict()
        for node in tree.postorder_node_iter():
            if node.taxon and node.taxon.label:
                new_node_id = pdm.add_node(taxon=node.taxon.label)
            else:
                new_node_id = pdm.add_node()
            node_to_id[node] = new_node_id

        for node in tree.postorder_node_iter():
            if node.parent_node is not None:
                pdm.add_edge(parent_id=node_to_id[node.parent_node],
                             child_id=node_to_id[node],
                             length=node.edge_length)
        return pdm

    def add_node(self, taxon: Optional[str] = None) -> int:
        """Add a new node to the tree.

        Args:
            taxon: The taxon name if this is a leaf node.

        Returns:
            The index of the new node.
        """
        return self._rs.add_node(taxon=taxon)

    def add_edge(self, parent_id: int, child_id: int, length: float):
        """Add an edge between the two nodes

        Args:
            parent_id: The index of the parent node.
            child_id: The index of the child node.
            length: The length of the edge.
        """
        return self._rs.add_edge(parent_id=parent_id, child_id=child_id, length=length)

    def update_edge_lengths(self, child_nodes: np.ndarray, new_edge_lengths: np.ndarray):
        return self._rs.update_edge_lengths(child_nodes=child_nodes, lengths=new_edge_lengths)
    
    def update_all_edge_lengths(self, length: float):
        return self._rs.update_all_edge_lengths(length=length)

    def get_nodes(self) -> List[int]:
        """Return all node indexes in the tree."""
        return self._rs.get_nodes()

    def dm(self, norm: Optional[bool] = False) -> np.ndarray:
        """Returns a symmetrical distance matrix.

        Args:
            norm: If True, the matrix is normalized by branch length.
        """
        return self._rs.dm(norm=norm)

    def taxa(self) -> List[str]:
        """Returns a list of all taxa within the tree."""
        return self._rs.taxa()

    def length(self) -> float:
        """Returns the total length of the tree (sum of branch lengths)."""
        return self._rs.length()

    def compute_row_vec(self):
        """Compute the row vector for the tree (required if not initialised from a Newick file)."""
        return self._rs.compute_row_vec()

    def distance(self, a: str, b: str, norm: Optional[bool] = False) -> float:
        """Compute the distance between two taxa.

        Args:
            a: The first taxon.
            b: The second taxon.
            norm: If the distance should be normalised by the sum of branch lengths.

        Returns:
            The distance between the two taxa.
        """
        return self._rs.distance(a=a, b=b, norm=norm)



import matplotlib.patheffects as PathEffects
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import typer
from scipy.optimize import curve_fit

plt.rcParams['font.sans-serif'] = "Arial"


def plot_joint_axis(path_results, path_out):
    df = pd.read_csv(path_results, sep="\t")

    df = df[(df['n_taxa'] >= 1000)]
    df['tot_min'] = df['tot_sec'] / 60
    df['max_mem_gb'] = df['max_mem'] / 1024 / 1024

    df_dendro = df[df['source'] == 'dendropy']
    df_phylodm = df[df['source'] == 'phylodm']
    df_loading = df[df['source'] == 'loading']

    loading_cpu = df_loading[['n_taxa', 'tot_min']].groupby("n_taxa").median()
    loading_mem = df_loading[['n_taxa', 'max_mem_gb']].groupby("n_taxa").median()

    d_loading_cpu_mins = loading_cpu['tot_min'].to_dict()
    d_loading_mem_gb = loading_mem['max_mem_gb'].to_dict()

    t = list(map(int, df_phylodm['n_taxa'].values))
    phylodm_cpu_vals = [max(row['tot_min'] - d_loading_cpu_mins[row['n_taxa']], 0)
                        for idx, row in df_phylodm[['n_taxa', 'tot_min']].iterrows()]
    dendropy_cpu_vals = [max(row['tot_min'] - d_loading_cpu_mins[row['n_taxa']], 0)
                         for idx, row in df_dendro[['n_taxa', 'tot_min']].iterrows()]

    phylodm_mem_vals = [max(row['max_mem_gb'] - d_loading_mem_gb[row['n_taxa']], 0)
                        for idx, row in df_phylodm[['n_taxa', 'max_mem_gb']].iterrows()]
    dendropy_mem_vals = [max(row['max_mem_gb'] - d_loading_mem_gb[row['n_taxa']], 0)
                         for idx, row in df_dendro[['n_taxa', 'max_mem_gb']].iterrows()]

    fig, ax1 = plt.subplots(figsize=(8, 6))

    color_dendro = '#3727B3'
    color_phylo = '#179A39'
    markersize = 4

    # Polyfit
    popt_phylo_cpu, pcov = curve_fit(func, t, phylodm_cpu_vals)
    print(f'PhyloDM (cpu): x^2 * {popt_phylo_cpu[0]}')

    popt_phylo_mem, pcov = curve_fit(func, t, phylodm_mem_vals)
    print(f'PhyloDM (mem): x^2 * {popt_phylo_mem[0]}')

    popt_dendro_cpu, pcov = curve_fit(func, t, dendropy_cpu_vals)
    print(f'DendroPy (cpu): x^2 * {popt_dendro_cpu[0]}')

    popt_dendro_mem, pcov = curve_fit(func, t, dendropy_mem_vals)
    print(f'DendroPy (mem): x^2 * {popt_dendro_mem[0]}')

    alpha = 1

    xp = np.linspace(1000, 30000, 1000)
    p1 = ax1.plot(xp, func(xp, *popt_phylo_cpu), '-', color=color_phylo, label=f'PhyloDM (minutes) = x\u00b2 {popt_phylo_cpu[0]:.2e}')
    p2 = ax1.plot(xp, func(xp, *popt_phylo_mem), '--', color=color_phylo, label=f'PhyloDM (GB) = x\u00b2 {popt_phylo_mem[0]:.2e}')
    p3 = ax1.plot(xp, func(xp, *popt_dendro_cpu), '-', color=color_dendro, label=f'DendroPy (minutes) = x\u00b2 {popt_dendro_cpu[0]:.2e}')
    p4 = ax1.plot(xp, func(xp, *popt_dendro_mem), '--', color=color_dendro, label=f'DendroPy (GB) = x\u00b2 {popt_dendro_mem[0]:.2e}')

    ax1.set_xlabel('n_taxa')
    ax1.set_ylabel('Time (minutes) / Memory (GB)')
    ax1.plot(t, phylodm_cpu_vals, 'o', markersize=markersize, color=color_phylo, label='PhyloDM (minutes)', alpha=alpha)
    ax1.plot(t, dendropy_cpu_vals, 'o', markersize=markersize, color=color_dendro, label='DendroPy (minutes)', alpha=alpha)
    ax1.grid()
    ax1.set_yscale('log')

    ax1.plot(t, phylodm_mem_vals, 'x', markersize=markersize, color=color_phylo, label='PhyloDM (GB)', alpha=alpha)
    ax1.plot(t, dendropy_mem_vals, 'x', markersize=markersize, color=color_dendro, label='DendroPy (GB)', alpha=alpha)

    ax1.grid(True, which="both", ls="--", c='gray', alpha=0.6)

    ylabels = [f'{x:g}' for x in ax1.get_yticks()]
    ax1.set_yticklabels(ylabels)

    font_size = 12

    # Set maximum values
    ax1_phylo_max_x = max(t)
    ax1_phylo_max_y = func(ax1_phylo_max_x, popt_phylo_cpu)[0]
    txt = ax1.text(ax1_phylo_max_x, ax1_phylo_max_y, f'{ax1_phylo_max_y * 60:.1f} seconds',
                   fontsize=font_size, color=color_phylo, va='bottom', ha='right')
    txt.set_path_effects([PathEffects.withStroke(linewidth=4, foreground='w')])

    ax1_dendro_max_x = max(t)
    ax1_dendro_max_y = func(ax1_dendro_max_x, popt_dendro_cpu)[0]
    txt = ax1.text(ax1_dendro_max_x, ax1_dendro_max_y - 50, f'{ax1_dendro_max_y / 60:.1f} hours',
                   fontsize=font_size, color=color_dendro, va='top', ha='right')
    txt.set_path_effects([PathEffects.withStroke(linewidth=4, foreground='w')])

    # Set maximum values
    ax2_phylo_max_x = max(t)
    ax2_phylo_max_y = func(ax2_phylo_max_x, popt_phylo_mem)[0]
    txt = ax1.text(ax2_phylo_max_x, ax2_phylo_max_y, f'{ax2_phylo_max_y:.1f} GB',
                   fontsize=font_size, color=color_phylo, va='bottom', ha='right')
    txt.set_path_effects([PathEffects.withStroke(linewidth=4, foreground='w')])

    ax2_dendro_max_x = max(t)
    ax2_dendro_max_y = func(ax2_dendro_max_x, popt_dendro_mem)[0]
    txt = ax1.text(ax2_dendro_max_x, ax2_dendro_max_y, f'{ax2_dendro_max_y:.1f} GB',
                   fontsize=font_size, color=color_dendro, va='bottom', ha='right')
    txt.set_path_effects([PathEffects.withStroke(linewidth=4, foreground='w')])

    # fig.tight_layout()

    plots = p1 + p2 + p3 + p4
    labs = [p.get_label() for p in plots]
    ax1.legend(plots, labs, loc=0)

    plt.title('PhyloDM vs DendroPy distance matrix construction resource usage')
    plt.rcParams['svg.fonttype'] = 'none'

    # plt.show()
    plt.savefig(path_out)


def func(x, a):
    return x * x * a


def print_eqn(x, y):
    cpu_polyfit = np.polyfit(np.log(x), y, 1)
    # cpu_f = np.poly1d(cpu_polyfit)
    print(f'y = {cpu_polyfit[0]} log(x) + {cpu_polyfit[1]}')
    return cpu_polyfit[0], cpu_polyfit[1]


def main(path_results: str, path_out: str):
    # plot_subplot(path_results, path_out)
    plot_joint_axis(path_results, path_out)


if __name__ == "__main__":
    typer.run(main)



import dendropy
import typer


def main(path_tree: str):
    tree = dendropy.Tree.get_from_path(path_tree, schema='newick')
    tree.phylogenetic_distance_matrix()


if __name__ == "__main__":
    typer.run(main)



import os
import re

import pandas as pd
import typer


def parse_trial_output(path):
    with open(path) as fh:
        output = fh.read()

    wall_time = re.search(r'.*Elapsed \(wall clock\) time \(h:mm:ss or m:ss\): (.+)', output).group(1)
    max_mem = int(re.search(r'.*Maximum resident set size \(kbytes\): (\d+)', output).group(1))
    exit_code = int(re.search(r'.*Exit status: (\d)', output).group(1))

    if exit_code != 0:
        print(output)
        raise Exception(f'Non zero exit code: {path}')

    split_time = wall_time.split(':')
    if len(split_time) == 3:
        hr_sec = int(split_time[0]) * 3600
        mm_sec = int(split_time[1]) * 60
        sec = float(split_time[2])
        tot_sec = hr_sec + mm_sec + sec
    else:
        mm_sec = int(split_time[0]) * 60
        sec = float(split_time[1])
        tot_sec = mm_sec + sec

    return tot_sec, max_mem


def get_trials_from_dir(dir_path: str, source: str):
    rows = list()
    for file in os.listdir(dir_path):
        if not file.endswith('.txt'):
            continue

        n_taxa, trial = file.split('_')
        n_taxa = int(n_taxa)
        trial = int(trial.split('.')[0])

        path = os.path.join(dir_path, file)
        tot_sec, max_mem = parse_trial_output(path)

        rows.append({
            'source': source,
            'n_taxa': n_taxa,
            'trial': trial,
            'tot_sec': tot_sec,
            'max_mem': max_mem
        })
    return pd.DataFrame(rows)


def main(dendropy_dir: str, phylodm_dir: str, loading_dir: str, output_path: str):
    df_dendropy = get_trials_from_dir(dendropy_dir, 'dendropy')
    df_phylodm = get_trials_from_dir(phylodm_dir, 'phylodm')
    df_loading = get_trials_from_dir(loading_dir, 'loading')

    df = pd.concat([df_phylodm, df_dendropy, df_loading], ignore_index=True)
    df.sort_values(by=['source', 'n_taxa', 'trial'], inplace=True)
    df.to_csv(output_path, index=False, sep='\t')


if __name__ == "__main__":
    typer.run(main)



import platform
import subprocess

import typer


def main(script_path: str, tree_path: str, out_path: str):
    exe = 'gtime' if platform.system() == 'Darwin' else '/usr/bin/time'
    args = [exe, '--verbose', 'python', script_path, tree_path]
    proc = subprocess.Popen(args, stderr=subprocess.PIPE, encoding='utf-8')
    stdout, stderr = proc.communicate()
    if proc.returncode == 0:
        with open(out_path, 'w') as fh:
            fh.write(stderr)
    else:
        raise Exception(f'Non-zero return code: {stderr}')
    return


if __name__ == "__main__":
    typer.run(main)
